import hashlib
import json
import logging
import os
import socket
import threading
import time
import uuid
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

from .util import constants as C
from .util import helper
from .util.decorators import *
from .util.enums import State
from .util.exceptions import FaktroyWorkProtocolError


class Client:
    """
    Faktory Client.

    This is an implementation of FWP (Faktory Work Protocol) client.

    FWP allows a client to interact with a Faktory work server. It permits
    a client to authenticate to a Faktory server, submit units of work (jobs)
    for later execution, and/or fetch units of work for processing and
    subsequently report their execution result.

    Parameters
    ----------
    faktory_url : Optional[str]
        URL of the Faktory server with which the client will establish
        a connection. Following format is expected: `tcp://:password@host:port`.
        If the url is not given, the client will try to find it
        in the environment variable `FAKTORY_URL`.
    role : {{'producer', 'consumer', 'both'}}, default 'both'
        The role of the client, some commands are only available for one role
        or another. For the client to act as both a consumer and a producer use
        `'both'`.
    timeout : Optional[int], default 30
        Timeout on blocking socket operations. If zero is given, the socket
        is put in non-blocking mode. If `None` is given, the socket is put
        in blocking mode.
    worker_id : Optional[str]
        Globally unique identifier for the worker that will use this client.
        If the client's role is `producer`, this argument is ignored.
    labels : List[str], default ['python']
        Labels that apply to the worker using this client. These labels
        will be displayed in Faktory webui.
        If the client's role is `producer`, this argument is ignored.
    beat_period : int, default 15
        The period in seconds for sending BEAT to the server to recognize state
        changes initiated by the server. Period must be between 5 and 60 seconds.

    Notes
    -----
    `FWP specification <https://github.com/contribsys/faktory/blob/master/docs/protocol-specification.md>`_
    """
    def __init__(self,
                 faktory_url: Optional[str] = None,
                 role: str = 'producer',
                 timeout: Optional[int] = 30,
                 worker_id: Optional[str] = None,
                 labels: List[str] = C.DEFAULT_LABELS,
                 beat_period: int = C.RECOMMENDED_BEAT_PERIOD) -> None:
        self.logger = logging.getLogger(name='FaktoryClient')

        if role not in ['consumer', 'producer', 'both']:
            raise ValueError(
                f"Unexpected role ({role}), role should be 'consumer', 'producer' or 'both'"
            )
        self.role = role

        if not faktory_url:
            faktory_url = os.environ.get('FAKTORY_URL', C.DEFAULT_FAKTORY_URL)

        parsed_url = urlparse(faktory_url)
        self.host = parsed_url.hostname
        self.port = parsed_url.port or C.DEFAULT_PORT
        self.password = parsed_url.password

        self.sock: socket.socket
        self.timeout = timeout
        self.state = State.DISCONNECTED

        self.rlock = threading.RLock()

        # Consumer specific fields
        if self.role != 'producer':
            self.labels = labels

            if worker_id is None:
                self.logger.warning(
                    f'No worker id has been given, a random id will be used')
            elif len(worker_id) < 8:
                raise ValueError(
                    'Worker id must be a string of at least 8 characters')

            self.worker_id = worker_id or uuid.uuid4().hex

            if not (C.MIN_ALLOOWABLE_BEAT_PERIOD <= beat_period <=
                    C.MAX_ALLOOWABLE_BEAT_PERIOD):
                ValueError(
                    'Beat period is {beat_period}, but should be between {C.MIN_ALLOOWABLE_BEAT_PERIOD} and {C.MAX_ALLOOWABLE_BEAT_PERIOD}'
                )
            self.beat_period = beat_period
            self.rss_kb = None
            self.heartbeat_thread: threading.Thread

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.disconnect()

    @valid_states_cmd([State.DISCONNECTED])
    def connect(self) -> bool:
        self.logger.info(f'Client lifecycle state is {self.state}')

        self.logger.info('Openning connection...')
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.sock.settimeout(self.timeout)
        self.sock.connect((self.host, self.port))

        # The very first message from server
        msg = self._receive()

        self._raise_error(msg)

        self.logger.info('Connection to server is successfully established')
        self._set_state(State.NOT_IDENTIFIED)

        msg_args = json.loads(msg[3:])
        protocol_version = int(msg_args["v"])
        if protocol_version != 2:
            raise FaktroyWorkProtocolError(
                f'Only FWP version 2 supported, got {protocol_version}')

        password_hash = None
        if 'i' in msg_args and 's' in msg_args:
            self.logger.info('Password is required')
            if not self.password:
                raise ValueError('Password required but not provided')
            password_hash_iterations = msg_args['i']
            password_hash_salt = msg_args['s']
            self.logger.info('Hashing password...')
            password_hash = str.encode(str(
                self.password)) + str.encode(password_hash_salt)
            for _ in range(password_hash_iterations):
                password_hash = hashlib.sha256(password_hash).digest()
        else:
            self.logger.info('Password is not required')

        if password_hash:
            _ = self._hello(pwdhash=password_hash.hex())
        else:
            _ = self._hello()

        return True

    def disconnect(self):
        if self.state != State.END:
            with self.rlock:
                self._end()

        self.logger.info("Disconnecting...")
        self.sock.close()
        self._set_state(State.DISCONNECTED)

    def _set_state(self, new_state: State) -> bool:
        if self.role == 'producer' and new_state in [
                State.QUIET, State.TERMINATING
        ]:
            raise FaktroyWorkProtocolError(
                f'Producer Client cannot enter {new_state} stage')

        old_state = self.state

        if old_state == new_state:
            self.logger.info(f'client state is {old_state}, not changed')
            return False

        self.state = new_state
        self.logger.info(
            f'Client state changed from {old_state} to {new_state}')

        if self.role != 'producer':
            # If client acts as consumer, start heartbeating
            # when IDENTIFIED state is entered
            if self.state == State.IDENTIFIED:
                self.heartbeat_thread = threading.Thread(
                    target=self._heartbeat, args=())
                self.heartbeat_thread.start()

        return True

    def _send(self, command: str):
        self.sock.send(command.encode('utf-8'))
        self.logger.debug(f'C: {command}')

    def _receive(self) -> str:
        msg = self.sock.recv(1024).decode('utf-8')

        while not helper.RESP.is_message_complete(msg):
            msg += self.sock.recv(1024).decode('utf-8')

        msg = msg.strip()
        self.logger.debug(f'S: {msg}')
        return msg

    def _send_and_receive(self, command: str) -> str:
        with self.rlock:
            self._send(command)
            msg = self._receive()
            return msg

    def _raise_error(self, faktory_response):
        if faktory_response[0] == '-':
            raise FaktroyWorkProtocolError(
                f'Error received from Faktory server: {faktory_response}')

    def _heartbeat(self):
        while self.state in [State.IDENTIFIED, State.QUIET]:
            with self.rlock:
                self.logger.info(
                    f'Sending heartbeat to server, next heartbeat in {self.beat_period} seconds'
                )
                self._beat(rss_kb=self.rss_kb)
            time.sleep(self.beat_period)

    ####
    ## Client commands
    ####

    @valid_states_cmd([State.NOT_IDENTIFIED])
    def _hello(self, pwdhash=None) -> bool:
        """
        The HELLO command MUST be the first command issued by any client when
        connecting to a Faktory server. It is sent in response to the server's
        initial HI message.
        """
        client_info: Dict[str, Any] = {
            # Protocol version number
            "v": 2,
        }

        # Required fields for consumers
        if self.role != 'producer':
            client_info.update({
                "hostname": self.host,
                "wid": self.worker_id,
                "pid": os.getpid(),
                "labels": self.labels,
            })

        # Required fields for protected server
        if pwdhash:
            client_info['pwdhash'] = pwdhash

        command = f'HELLO {json.dumps(client_info)}{C.CRLF}'
        msg = self._send_and_receive(command)

        self._raise_error(msg)
        self._set_state(State.IDENTIFIED)

        return True

    @valid_states_cmd([State.IDENTIFIED])
    def _flush(self) -> bool:
        """
        FLUSH allows to clear all info from Faktory's internal database.
        It uses Redis's FLUSHDB command under the covers.
        """
        command = f'FLUSH{C.CRLF}'
        msg = self._send_and_receive(command)
        self._raise_error(msg)
        return True

    @valid_states_cmd([State.IDENTIFIED])
    def _info(self) -> str:
        command = f'INFO{C.CRLF}'
        msg = self._send_and_receive(command)
        self._raise_error(msg)
        return msg

    @valid_states_cmd([State.IDENTIFIED, State.QUIET, State.TERMINATING])
    def _end(self):
        """
        The END command is used to signal to the server that it wishes
        to terminate the connection.
        """
        command = f'END{C.CRLF}'
        self._send(command)
        self._set_state(State.END)

    ####
    ## Producer commands
    ####

    @producer_cmd
    @valid_states_cmd([State.IDENTIFIED])
    def _push(self, work_unit: Dict) -> bool:
        command = f'PUSH {json.dumps(work_unit)}{C.CRLF}'
        msg = self._send_and_receive(command)
        self._raise_error(msg)
        return True

    ####
    ## Consumer commands
    ####

    @consumer_cmd
    @valid_states_cmd([State.IDENTIFIED])
    def _fetch(self, queues: List[str] = []) -> Optional[Dict]:
        command = f'FETCH {" ".join(queues)}{C.CRLF}'
        msg = self._send_and_receive(command)

        self._raise_error(msg)

        n_bytes, data = helper.RESP.parse_bulk_string(msg)

        if n_bytes == -1:
            return None
        else:
            return json.loads(data)

    @consumer_cmd
    @valid_states_cmd([State.IDENTIFIED, State.QUIET, State.TERMINATING])
    def _ack(self, jid: str) -> bool:
        args = {'jid': jid}
        command = f'ACK {json.dumps(args)}{C.CRLF}'
        msg = self._send_and_receive(command)
        self._raise_error(msg)
        return True

    @consumer_cmd
    @valid_states_cmd([State.IDENTIFIED, State.QUIET, State.TERMINATING])
    def _fail(self, jid: str, errtype: str, message: str,
              backtrace: List[str]) -> bool:
        args = {
            'jid': jid,
            'errtype': errtype,
            'message': message,
            'backtrace': backtrace,
        }
        command = f'FAIL {json.dumps(args)}{C.CRLF}'
        msg = self._send_and_receive(command)
        self._raise_error(msg)
        return True

    @consumer_cmd
    @valid_states_cmd([State.IDENTIFIED, State.QUIET])
    def _beat(self, rss_kb: Optional[int] = None) -> bool:
        args: Dict[str, Any] = {
            'wid': self.worker_id,
        }

        if self.state == State.QUIET:
            args['current_state'] = 'quiet'
        # TODO: check if this meet FWP
        elif self.state == State.TERMINATING:
            args['current_state'] = 'terminate'

        if rss_kb:
            args['rss_kb'] = rss_kb

        command = f'BEAT {json.dumps(args)}{C.CRLF}'
        msg = self._send_and_receive(command)

        self._raise_error(msg)

        if msg[0] == '+':
            if msg[1:3] == 'OK':
                return True
        # Bulk String
        elif msg[0] == '$':
            _, data = helper.RESP.parse_bulk_string(msg)
            data = json.loads(data)
            if data['state'] == 'quiet':
                self._set_state(State.QUIET)
            elif data['state'] == 'terminate':
                self._set_state(State.TERMINATING)
            else:
                raise FaktroyWorkProtocolError(
                    f'Unexpected BEAT response state ({data})')
        else:
            raise FaktroyWorkProtocolError(
                f'Unexpected BEAT response: ({msg})')

        return True



import uuid
from typing import Any, Dict, List, Optional

from .client import Client


class Producer:
    """
    Faktory Producer.

    Parameters
    ----------
    client : pyfaktory.Client
        Faktory client whose `role` must be either 'producer' or 'both'.
    """
    def __init__(self, client: Client) -> None:
        if client.role == 'consumer':
            raise ValueError(
                "Provided client is exclusively consumer and can't act as a producer"
            )
        self.client = client

    def push_job(self,
                 jobtype: str,
                 jid: Optional[str] = None,
                 args: List[Any] = [],
                 queue: str = 'default',
                 reserve_for: int = 1800,
                 at: str = '',
                 retry: int = 25,
                 backtrace: int = 5,
                 custom: Optional[Dict] = None) -> bool:
        """
        Push a job to Faktory server.

        Parameters
        ----------
        jid : str
            Globally unique ID for the job.
        jobtype : str
            Discriminator used by a worker to decide how to execute a job.
        args : List[Any]
            Parameters the worker should use when executing the job.
        queue : str, default 'default'
            Which job queue to push this job onto.
        reserve_for : int, default 1800
            Number of seconds a job may be held by a worker before
            it is considered failed (must be greater than 60).
        at : str, default ''
            Run the job at approximately this time, immediately if blank.
            RFC3339 string is expected, but no validation is done client side.
        retry : int, default 25
            Number of times to retry this job if it fails.
            0 discards the failed job, -1 saves the failed
            job to the dead set.
        backtrace : int, default 5
            Number of lines of FAIL information to preserve.
        custom : dict, default None
            Provides additional context to the worker executing the job.

        Returns
        -------
        bool
            Indicates if the push was successful.
        """
        if not jid:
            jid = uuid.uuid4().hex

        # Mandatory fields
        job: Dict[str, Any] = {
            'jid': jid,
            'jobtype': jobtype,
            'args': args,
        }

        if reserve_for < 60:
            raise ValueError(
                f'`reserve_for` == {reserve_for}, but must be greater than 60')

        # TODO: add `at` validation

        # Optional fields
        job.update({
            'queue': queue,
            'reserve_for': reserve_for,
            'at': at,
            'retry': retry,
            'backtrace': backtrace,
            'custom': custom,
        })

        # Optional fields can be omitted for some values.
        # TODO: omit fields when possible (this does not change anything,
        # it just allows to reduce the size of the command to send)

        return self.client._push(job)



class FaktoryError(Exception):
    """
    Base Faktory Exception.
    """
    pass


class FaktroyWorkProtocolError(FaktoryError):
    """
    Faktroy Work Protocol Error.
    """
    pass


class FaktoryConsumerError(FaktoryError):
    """
    Faktory Worker Error.
    """
    pass



import random

from . import constants as C


class RESP:
    """
    RESP (REdis Serialization Protocol).

    Notes
    -----
    `RESP <https://redis.io/topics/protocol>`_
    """
    @staticmethod
    def parse_bulk_string(s: str):
        assert s[0] == '$'

        try:
            n_bytes, data = s[1:].split('\r\n')
            return int(n_bytes), data
        except Exception:
            return -1, ''

    @staticmethod
    def is_message_complete(msg: str):
        if len(msg) < 2:
            return False

        if not msg.endswith(C.CRLF):
            return False

        if msg[0] == '$' and msg[1] != '-':
            nb_crlf = 2
        else:
            nb_crlf = 1

        return msg.count(C.CRLF) == nb_crlf


# http://utopia.duth.gr/~pefraimi/research/data/2007EncOfAlg.pdf
def weighted_shuffle(items, weights):
    order = sorted(range(len(items)),
                   key=lambda i: random.random()**weights[i])
    return [items[i] for i in order]



from .client import Client
from .consumer import Consumer
from .producer import Producer
from .util.exceptions import *

__version__ = "0.1.0"
__url__ = "https://github.com/ghilesmeddour/faktory_worker_python"



from functools import wraps
from typing import List

from .enums import State
from .exceptions import FaktroyWorkProtocolError


def producer_cmd(fn):
    @wraps(fn)
    def wrapper(self, *args, **kw):
        if self.role == 'consumer':
            raise FaktroyWorkProtocolError(
                f'Trying to send producer command `{fn.__name__}` with a consumer client'
            )
        return fn(self, *args, **kw)

    return wrapper


def consumer_cmd(fn):
    @wraps(fn)
    def wrapper(self, *args, **kw):
        if self.role == 'producer':
            raise FaktroyWorkProtocolError(
                f'Trying to send consumer command `{fn.__name__}` with a producer client'
            )
        return fn(self, *args, **kw)

    return wrapper


def valid_states_cmd(states: List[State]):
    def decorator(fn):
        @wraps(fn)
        def wrapper(self, *args, **kw):
            if self.state not in states:
                raise FaktroyWorkProtocolError(
                    f'Client state is {self.state} while `{fn.__name__}` is only valid in the states {states}'
                )
            return fn(self, *args, **kw)

        return wrapper

    return decorator



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfaktory",
    version="0.1.0",
    author="Ghiles Meddour",
    author_email="ghiles.meddour@munic.io",
    description="Faktory Client Python (Producer and Consumer/Worker)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ghilesmeddour/faktory_worker_python",
    project_urls={
        "Bug Tracker":
        "https://github.com/ghilesmeddour/faktory_worker_python/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=[
        "pebble",
    ],
)



import logging
import multiprocessing
import random
import signal
import sys
import time
import traceback
from typing import Callable, Dict, List, Optional

from pebble import ProcessPool, sighandler

from .client import Client
from .util import constants as C
from .util import helper
from .util.enums import State


class Consumer:
    """
    Faktory Consumer (Worker).

    It fetches units of work (jobs) from the server and executes them.
    It retrieves the jobs from the queues, and decides how to execute them
    based on the jobtype.

    A handler must be attached to each jobtype before the consumer is launched.

    Parameters
    ----------
    client : pyfaktory.Client
        Faktory client. Client `role` must be either 'consumer' or 'both'.
    queues : List[str], default ['default']
        The queues from which the consumer will fetch jobs. If you provide
        no `queues`, the consumer will process `default` queue.
    priority : {{'strict', 'uniform', 'weighted'}}, default 'uniform'
        Priority indicates in which queue order the jobs should be fetched
        first. With `strict` priority, the worker always fetches from the first
        queue and will only try to fetch from the next once the previous queue
        is empty. With `uniform`, each queue has an equal chance of being
        fetched first. With `weighted`, queues have a different probability
        of being fetched. These probabilities are specified with `weights`
        argument.
    weights : Optional[List[float]], default None
        Probability of the queues to be fetched. This parameter is required
        when `priority` is `weighted` (and ignored in other cases), and must
        have the same number of elements as `queues`.
    concurrency : int, default 4
        Number of jobs to run at the same time.
    grace_period : int, default 25
        Grace period between the beginning of a shutdown and its end.
        This period is used to give the job some time to finish, to stop them
        properly and to notify the server. This period should never be longer
        than 30 seconds.
    """
    def __init__(self,
                 client: Client,
                 queues: List[str] = C.DEFAULT_QUEUES,
                 priority: str = 'uniform',
                 weights: Optional[List[float]] = None,
                 concurrency: int = 4,
                 grace_period: int = C.DEFAULT_GRACE_PERIOD) -> None:
        self.logger = logging.getLogger(name='FaktoryConsumer')

        if client.role == 'producer':
            raise ValueError(
                "Provided client is exclusively producer and can't act as a consumer"
            )
        self.client = client

        self.queues = queues

        if priority not in ['strict', 'uniform', 'weighted']:
            raise ValueError(
                f"Unexpected priority ({priority}), priority should be 'strict', 'uniform' or 'weighted'"
            )
        self.priority = priority

        if self.priority == 'weighted':
            if weights is None:
                raise ValueError(
                    'Priority is weighted but weights are not provided')
            elif len(weights) != len(self.queues):
                raise ValueError('Weights and queues lengths mismatch')
            else:
                self.weights = weights

        self.concurrency = concurrency
        self.grace_period = grace_period

        self.pool = ProcessPool(max_workers=self.concurrency)

        self.job_handlers: Dict[str, Callable] = {}

        self.pending_tasks_count = 0
        self.lock_pending_tasks_count = multiprocessing.Lock()

    @sighandler((signal.SIGTERM))
    def handle_sigterm(*_):
        raise KeyboardInterrupt

    def register(self, name: str, fn: Callable):
        """
        Register a handler for the given jobtype.

        It is expected that all jobtypes are registered upon process
        startup.
        """
        self.job_handlers[name] = fn
        self.logger.info(f"Registered handler for jobtype: {name}")

    def get_job_handler(self, name: str) -> Callable:
        try:
            return self.job_handlers[name]
        except KeyError:
            self.logger.error(f"'{name}' is not a registered job handler")
            # One could consider just failing the job and continue running,
            # but we are not doing it here because it is expected that all
            # jobtypes are registered upon process startup.
            raise ValueError(f"'{name}' has no registered handler")

    def get_queues(self) -> List[str]:
        if self.priority == 'strict':
            return self.queues
        elif self.priority == 'uniform':
            random.shuffle(self.queues)
            return self.queues
        else:
            return helper.weighted_shuffle(self.queues, self.weights)

    def task_done(self, future):
        self.logger.info(f'Task done callback called for job {future.job_id}')
        try:
            result = future.result()
            self.logger.info(f'Task (job {future.job_id}) returned {result}')
            self.client._ack(jid=future.job_id)
        except Exception:
            err_type, err_value, err_traceback = sys.exc_info()
            self.logger.info(
                f'Task (job {future.job_id}) raised {err_type}: {err_value}')
            self.client._fail(jid=future.job_id,
                              errtype=err_type.__name__,
                              message=str(err_value),
                              backtrace=traceback.format_tb(
                                  err_traceback, limit=future.backtrace))
        finally:
            with self.lock_pending_tasks_count:
                self.pending_tasks_count -= 1

    def run(self):
        """
        Start the consumer.

        When this method is called, the fetching and execution of the jobs
        starts. The job handlers must have been registered beforehand.

        This method is blocking, it only stops in the event of an error
        (only main loop errors, errors that occur in the handlers cause the job
        to fail and are reported to Faktory Server) or when a signal is received
        (Ctrl-C or from the Faktory Web UI).

        At the beginning of the shutdown, the worker gives itself a grace period
        to stop properly and notify the last information to the Faktory server.
        If a second signal is received, this causes an immediate shutdown.
        """
        self.logger.info('Entering run loop..')
        # TODO: check this
        while self.client.state in [State.IDENTIFIED, State.QUIET]:
            try:
                if self.client.state == State.QUIET:
                    self.logger.info(
                        f'State is {self.client.state}, not fetching further jobs'
                    )
                    time.sleep(self.client.beat_period)
                    continue

                if self.pending_tasks_count < self.concurrency:
                    queues_tmp = self.get_queues()
                    self.logger.info(f'Fetching from queues: {queues_tmp}')
                    # If no jobs are found, _fatch will block for up
                    # to 2 seconds on the first queue provided.
                    job = self.client._fetch(queues_tmp)

                    if job:
                        # TODO: check
                        # timeout=job.get('reserve_for', None)
                        job_handler = self.get_job_handler(job.get('jobtype'))

                        future = self.pool.schedule(job_handler,
                                                    args=job.get('args'))

                        with self.lock_pending_tasks_count:
                            self.pending_tasks_count += 1

                        future.job_id = job.get("jid")
                        future.backtrace = job.get("backtrace", 0)
                        future.add_done_callback(self.task_done)
                    else:
                        self.logger.debug('Queues are empty.')
                else:
                    # TODO: maybe use Event object instead of sleep
                    time.sleep(0.1)
            except KeyboardInterrupt:
                self.logger.info(
                    'First KeyboardInterrupt, stopping after grace period')
                break
            except Exception:
                self.logger.error(
                    f'Shutting down due to an error: {traceback.format_exc()}')
                break

        self.logger.info(f'Run loop exited, state is {self.client.state}')
        self.logger.info(f'Grace period of {self.grace_period} seconds...')
        self.logger.info(f'Press Ctrl-C again to stop immediately')

        try:
            self.pool.close()
            self.pool.join(timeout=self.grace_period)
        except KeyboardInterrupt:
            self.logger.info('Second KeyboardInterrupt, stopping immediately')

        self.logger.info(f'End of the grace period. Stopping.')
        sys.exit(1)


# TODO: set rss_kb (sys.getsizeof?)



####
## Client
####

CRLF = '\r\n'
DEFAULT_FAKTORY_URL = 'tcp://localhost:7419'
DEFAULT_PORT = 7419
DEFAULT_LABELS = ['python']
DEFAULT_QUEUES = ['default']

####
## Producer
####

RECOMMENDED_BEAT_PERIOD = 15  # seconds
MIN_ALLOOWABLE_BEAT_PERIOD = 5  # seconds
MAX_ALLOOWABLE_BEAT_PERIOD = 60  # seconds

####
## Consumer
####

DEFAULT_GRACE_PERIOD = 25
MAX_GRACE_PERIOD = 30
